#include "TDateBook.h"
#include <stdio.h>
#include <fstream>
#include <time.h>

#pragma warning(disable : 4996)

namespace PalmDateBook {

bool IsCPUBigEndian() {
	return *((short *)"01") == 0x3031;
}

bool Read8Bit(std::istream *fs, unsigned char *data) {
	if (fs->eof())
		return false;
	int d = fs->get();
	*data = (d & 0xFF);
	return true;
}

bool Read16Bit(std::istream *fs, unsigned short *data) {
	unsigned char c[2];
	for (int i = 0; i < 2; i++) {
		if (!Read8Bit(fs, &c[i]))
			return false;
	}
	if (IsCPUBigEndian())
		*data = (c[1] | (c[0] << 8));
	else
		*data = (c[0] | (c[1] << 8));
	return true;
}

bool Read32Bit(std::istream *fs, unsigned int *data) {
	unsigned char c[4];
	for (int i = 0; i < 4; i++) {
		if (!Read8Bit(fs, &c[i]))
			return false;
	}
	if (IsCPUBigEndian())
		*data = (c[3] | (c[2] << 8) | (c[1] << 16) | (c[0] << 24));
	else
		*data = (c[0] | (c[1] << 8) | (c[2] << 16) | (c[3] << 24));
	return true;
}

bool Write8Bit(std::ostream *fs, unsigned char data) {
	fs->put(data);
	return true;
}

bool Write16Bit(std::ostream *fs, unsigned short data) {
	unsigned char c[2];
	if (IsCPUBigEndian()) {
		c[1] = data & 0xFF;
		c[0] = (data >> 8) & 0xFF;
	}
	else {
		c[0] = data & 0xFF;
		c[1] = (data >> 8) & 0xFF;
	}
	for (int i = 0; i < 2; i++) {
		if (!Write8Bit(fs, c[i]))
			return false;
	}
	return true;
}

bool Write32Bit(std::ostream *fs, unsigned int data) {
	unsigned char c[4];
	if (IsCPUBigEndian()) {
		c[3] = data & 0xFF;
		c[2] = (data >> 8) & 0xFF;
		c[1] = (data >> 16) & 0xFF;
		c[0] = (data >> 24) & 0xFF;
	}
	else {
		c[0] = data & 0xFF;
		c[1] = (data >> 8) & 0xFF;
		c[2] = (data >> 16) & 0xFF;
		c[3] = (data >> 24) & 0xFF;
	}
	for (int i = 0; i < 4; i++) {
		if (!Write8Bit(fs, c[i]))
			return false;
	}
	return true;
}

bool TFieldType::Read(std::istream* fs) {
	return Read32Bit(fs, &m_fieldType);
}

bool TFieldType::Write(std::ostream* fs) {
	return Write32Bit(fs, m_fieldType);
}

bool TField::Read(std::istream* fs) {
	if (!TFieldType::Read(fs))
		return false;
	return Read32Bit(fs, &m_data);
}

bool TField::Write(std::ostream* fs) {
	if (!TFieldType::Write(fs))
		return false;
	return Write32Bit(fs, m_data);
}

bool TCString::Read(std::istream *fs) {
	unsigned char slength;
	if (!Read8Bit(fs, &slength))
		return false;
	unsigned short length;
	if (slength == 0xFF) {
		if (!Read16Bit(fs, &length))
			return false;
	}
	else {
		length = slength;
	}
	unsigned char *bf = new unsigned char[length + 1];
	for (unsigned short i = 0; i < length; i++) {
		if (!Read8Bit(fs, &bf[i]))
			return false;
	}
	bf[length] = 0;
	m_data = (char *)bf;
	delete [] bf;
	return true;
}

bool TCString::Write(std::ostream *fs) {
	unsigned short length = (unsigned short)m_data.size();
	unsigned char slength;
	if (length < 0xFF) {
		slength = (unsigned char)length;
		if (!Write8Bit(fs, slength))
			return false;
	}
	else {
		if (!Write8Bit(fs, 0xFF))
			return false;
		if (!Write16Bit(fs, length))
			return false;
	}
	for (unsigned short i = 0; i < length; i++) {
		if (!Write8Bit(fs, m_data[i]))
			return false;
	}
	return true;
}

int TCString::CountLF() const {
	int count = 0;
	for (int i = 0; i < Length(); i++) {
		if (m_data[i] == 0x0A)
			count++;
	}
	return count;
}

bool TCString::BiggerThan(const TCString& s) const {
	if (CountLF() > s.CountLF())
		return true;
	if (Length() > s.Length())
		return true;
	return false;
}

bool TCStringField::Read(std::istream *fs) {
	if (!TField::Read(fs))
		return false;
	return m_text.Read(fs);
}

bool TCStringField::Write(std::ostream *fs) {
	if (!TField::Write(fs))
		return false;
	return m_text.Write(fs);
}

bool TClassEntry::Read(std::istream *fs) {
	if (!Read16Bit(fs, &m_constant))
		return false;
	unsigned short length;
	if (!Read16Bit(fs, &length))
		return false;
	unsigned char *bf = new unsigned char[length + 1];
	for (unsigned short i = 0; i < length; i++) {
		if (!Read8Bit(fs, &bf[i]))
			return false;
	}
	bf[length] = 0;
	m_className = (char *)bf;
	delete [] bf;
	return true;
}

bool TClassEntry::Write(std::ostream *fs) {
	if (!Write16Bit(fs, m_constant))
		return false;
	unsigned short length = (unsigned short)m_className.size();
	if (!Write16Bit(fs, length))
		return false;
	for (unsigned short i = 0; i < length; i++) {
		if (!Write8Bit(fs, m_className[i]))
			return false;
	}
	return true;
}

bool TRepeatEvent::Read(std::istream *fs) {
	if (!TFieldType::Read(fs))
		return false;
	if (!Read16Bit(fs, &m_dateExceptions))
		return false;
	for (unsigned short i = 0; i < m_dateExceptions; i++) {
		unsigned int d;
		if (!Read32Bit(fs, &d))
			return false;
		m_exceptionEntry.push_back(d);
	}
	if (!Read16Bit(fs, &m_repeatEventFlag))
		return false;
	if (m_repeatEventFlag == 0)
		return true;
	if (m_repeatEventFlag == 0xFFFF) {
		if (!m_classEntry.Read(fs))
			return false;
	}
	if (!Read32Bit(fs, &m_brand))
		return false;
	if (!Read32Bit(fs, &m_interval))
		return false;
	if (!Read32Bit(fs, &m_endDate))
		return false;
	if (!Read32Bit(fs, &m_firstDayOfWeek))
		return false;

	if (m_brand == REPEAT_BRABD_DAILY) {
		if (!Read32Bit(fs, &m_dayIndex))
			return false;
	}
	else if (m_brand == REPEAT_BRABD_WEEKLY) {
		if (!Read32Bit(fs, &m_dayIndex))
			return false;
		if (!Read8Bit(fs, &m_dayMask))
			return false;
	}
	else if (m_brand == REPEAT_BRABD_MONTHLYBYDAY) {
		if (!Read32Bit(fs, &m_dayIndex))
			return false;
		if (!Read32Bit(fs, &m_weekIndex))
			return false;
	}
	else if (m_brand == REPEAT_BRABD_MONTHLYBYDATE) {
		if (!Read32Bit(fs, &m_dayNumber))
			return false;
	}
	else if (m_brand == REPEAT_BRABD_YEARLYBYDATE) {
		if (!Read32Bit(fs, &m_dayNumber))
			return false;
		if (!Read32Bit(fs, &m_monthIndex))
			return false;
	}
	return true;
}

bool TRepeatEvent::Write(std::ostream *fs) {
	if (!TFieldType::Write(fs))
		return false;
	if (!Write16Bit(fs, m_dateExceptions))
		return false;
	for (unsigned short i = 0; i < m_dateExceptions; i++) {
		if (!Write32Bit(fs, m_exceptionEntry[i]))
			return false;
	}
	if (!Write16Bit(fs, m_repeatEventFlag))
		return false;
	if (m_repeatEventFlag == 0)
		return true;
	if (m_repeatEventFlag == 0xFFFF) {
		if (!m_classEntry.Write(fs))
			return false;
	}
	if (!Write32Bit(fs, m_brand))
		return false;
	if (!Write32Bit(fs, m_interval))
		return false;
	if (!Write32Bit(fs, m_endDate))
		return false;
	if (!Write32Bit(fs, m_firstDayOfWeek))
		return false;

	if (m_brand == REPEAT_BRABD_DAILY) {
		if (!Write32Bit(fs, m_dayIndex))
			return false;
	}
	else if (m_brand == REPEAT_BRABD_WEEKLY) {
		if (!Write32Bit(fs, m_dayIndex))
			return false;
		if (!Write8Bit(fs, m_dayMask))
			return false;
	}
	else if (m_brand == REPEAT_BRABD_MONTHLYBYDAY) {
		if (!Write32Bit(fs, m_dayIndex))
			return false;
		if (!Write32Bit(fs, m_weekIndex))
			return false;
	}
	else if (m_brand == REPEAT_BRABD_MONTHLYBYDATE) {
		if (!Write32Bit(fs, m_dayNumber))
			return false;
	}
	else if (m_brand == REPEAT_BRABD_YEARLYBYDATE) {
		if (!Write32Bit(fs, m_dayNumber))
			return false;
		if (!Write32Bit(fs, m_monthIndex))
			return false;
	}
	return true;
}

bool TDateBookEntry::Read(std::istream *fs) {
	if (!m_recordID.Read(fs))
		return false;
	if (!m_statusField.Read(fs))
		return false;
	if (!m_position.Read(fs))
		return false;
	if (!m_startTime.Read(fs))
		return false;
	if (!m_endTime.Read(fs))
		return false;
	if (!m_description.Read(fs))
		return false;
	if (!m_duration.Read(fs))
		return false;
	if (!m_note.Read(fs))
		return false;
	if (!m_untimed.Read(fs))
		return false;
	if (!m_private.Read(fs))
		return false;
	if (!m_category.Read(fs))
		return false;
	if (!m_alarmSet.Read(fs))
		return false;
	if (!m_alarmAdvUnits.Read(fs))
		return false;
	if (!m_alarmAdvType.Read(fs))
		return false;
	if (!m_repeatEvent.Read(fs))
		return false;
	return true;
}

bool TDateBookEntry::Write(std::ostream *fs) {
	if (!m_recordID.Write(fs))
		return false;
	if (!m_statusField.Write(fs))
		return false;
	if (!m_position.Write(fs))
		return false;
	if (!m_startTime.Write(fs))
		return false;
	if (!m_endTime.Write(fs))
		return false;
	if (!m_description.Write(fs))
		return false;
	if (!m_duration.Write(fs))
		return false;
	if (!m_note.Write(fs))
		return false;
	if (!m_untimed.Write(fs))
		return false;
	if (!m_private.Write(fs))
		return false;
	if (!m_category.Write(fs))
		return false;
	if (!m_alarmSet.Write(fs))
		return false;
	if (!m_alarmAdvUnits.Write(fs))
		return false;
	if (!m_alarmAdvType.Write(fs))
		return false;
	if (!m_repeatEvent.Write(fs))
		return false;
	return true;
}

std::string TDateBookEntry::GetSummary(VEVENT_ENCODE encode) {
	time_t tt = m_startTime.m_data;
	//struct tm *ptm = gmtime(&tt);
	struct tm *ptm = localtime(&tt);
	if (ptm == NULL) {
		std::cerr << "Time Error\n";
		return "Time Error";
	}
	char dateBuffer[16];
	char timeBuffer[16];
	sprintf(dateBuffer, "%04d-%02d-%02d", ptm->tm_year + 1900, ptm->tm_mon + 1, ptm->tm_mday);
	if (m_untimed.m_data)
		strcpy(timeBuffer, "*    ");
	else
		sprintf(timeBuffer, "%02d:%02d", ptm->tm_hour, ptm->tm_min);
	std::string result = dateBuffer;
	result += " ";
	result += timeBuffer;
	result += " ";
	if (encode == VEVENT_ENCODE_MIME_QP) {
		result += TUtil::MIMEEncode(m_description.m_text.m_data);
	}
	else /*if (encode == VEVENT_ENCODE_BACKSLASH_ESCAPE)*/ {
		result += TUtil::BackslashEscapeEncode(m_description.m_text.m_data);
	}
	return result;
}

bool TDateBookEntry::GetVEVENT(std::vector<std::string>* lines, bool bUTC, VEVENT_ENCODE encode) {
	time_t tt = m_startTime.m_data;
	struct tm *ptm = NULL;
	if (bUTC)
		ptm = gmtime(&tt);
	else
		ptm = localtime(&tt);
	if (ptm == NULL) {
		std::cerr << "Time Error\n";
		return false;
	}
	char dateBuffer[32];
	char timeBuffer[32];
	std::string s, smime;

	lines->push_back("BEGIN:VEVENT");
	s = GetVEVENT_MakeEncodedLine(m_description.m_text.m_data, "SUMMARY", encode);
	//smime = TUtil::MIMEEncode(m_description.m_text.m_data);
	//s = (smime != m_description.m_text.m_data) ? "SUMMARY;QUOTED-PRINTABLE:" : "SUMMARY:";
	//s += smime;
	lines->push_back(s);
	//
	if (m_note.m_text.m_data.length() > 0) {
		s = GetVEVENT_MakeEncodedLine(m_note.m_text.m_data, "DESCRIPTION", encode);
		//smime = TUtil::MIMEEncode(m_note.m_text.m_data);
		//s = (smime != m_note.m_text.m_data) ? "DESCRIPTION;QUOTED-PRINTABLE:" : "DESCRIPTION:";
		//s += smime;
		lines->push_back(s);
	}
	//
	sprintf(dateBuffer, "%04d%02d%02d", ptm->tm_year + 1900, ptm->tm_mon + 1, ptm->tm_mday);
	if (m_untimed.m_data && !bUTC) {
		timeBuffer[0] = 0;
	}
	else {
		sprintf(timeBuffer, "T%02d%02d%02d", ptm->tm_hour, ptm->tm_min, ptm->tm_sec);
	}
	s = "DTSTART:" + std::string(dateBuffer) + std::string(timeBuffer);
	if (bUTC)
		s += "Z";
	lines->push_back(s);
	//
	if (m_untimed.m_data) {
		tt += (60 * 60 * 24);
	}
	else {
		tt = m_endTime.m_data;
	}
	if (bUTC)
		ptm = gmtime(&tt);
	else
		ptm = localtime(&tt);
	if (ptm == NULL) {
		std::cerr << "Time Error\n";
		return false;
	}
	sprintf(dateBuffer, "%04d%02d%02d", ptm->tm_year + 1900, ptm->tm_mon + 1, ptm->tm_mday);
	if (m_untimed.m_data && !bUTC) {
		timeBuffer[0] = 0;
	}
	else {
		sprintf(timeBuffer, "T%02d%02d%02d", ptm->tm_hour, ptm->tm_min, ptm->tm_sec);
	}
	s = "DTEND:" + std::string(dateBuffer) + std::string(timeBuffer);
	if (bUTC)
		s += "Z";
	lines->push_back(s);
	//
	if (m_repeatEvent.GetRRULE(&s))
		lines->push_back(s);
	//
	lines->push_back("END:VEVENT");
	return true;
};

std::string TDateBookEntry::GetVEVENT_MakeEncodedLine(std::string data, std::string tag, VEVENT_ENCODE encode) {
	std::string s, s2, prop;
	if (encode == VEVENT_ENCODE_MIME_QP) {
		s2 = TUtil::MIMEEncode(data);
		prop = (s2 != data) ? ";QUOTED-PRINTABLE:" : ":";
	}
	else /*if (encode == VEVENT_ENCODE_BACKSLASH_ESCAPE)*/ {
		s2 = TUtil::BackslashEscapeEncode(data);
		prop = ":";
	}
	s = tag + prop + s2;
	return s;
}

bool TRepeatEvent::GetRRULE(std::string *outStr) {
	if (this->m_repeatEventFlag == 0)
		return false;
	time_t tt;
	struct tm *ptm = NULL;
	char dateBuffer[32];
	if (m_brand == REPEAT_BRABD_DAILY) {
		char *brand = "D";
		tt = m_endDate;
		ptm = localtime(&tt);
		if (ptm == NULL) {
			std::cerr << "Time Error\n";
			return false;
		}
		sprintf(dateBuffer, "%s%d %04d%02d%02d", brand, m_interval, ptm->tm_year + 1900, ptm->tm_mon + 1, ptm->tm_mday);
		*outStr = "RRULE:" + std::string(dateBuffer);
		return true;
	}
	else {
		std::cerr << "RRULE Alert\n";
		*outStr = "COMMENT:#RRULE ALERT";
		return true;
	}
	return false;
}

// ߂ČIQUOTED-PRINTABLEMIMEGR[hsB
std::string TUtil::MIMEEncode(std::string inStr) {
	std::string outStr;
	int count = (int)inStr.length();
	for (int i = 0; i < count; i++) {
		unsigned char c = inStr[i];
		if (c < ' ' || c == 0x7F || c == '=') {
			char buff[8];
			sprintf(buff, "=%02X", c);
			outStr += std::string(buff);
		}
		else {
			outStr += c;
		}
	}
	return outStr;
}

// ȑfȃobNXbVGXP[ṽGR[hsB
std::string TUtil::BackslashEscapeEncode(std::string inStr) {
	std::string outStr;
	int count = (int)inStr.length();
	for (int i = 0; i < count; i++) {
		unsigned char c = inStr[i];
		if (c == '\n') {
			outStr += "\\n";
		}
		else if (c == ';' || c == ',') {
			outStr += "\\";
			outStr += c;
		}
		else if (c < ' ' || c == 0x7F) {
			// ̂ĂB
		}
		else {
			outStr += c;
		}
	}
	return outStr;
}

bool TCategoryEntry::Read(std::istream *fs) {
	if (!Read32Bit(fs, &m_index))
		return false;
	if (!Read32Bit(fs, &m_id))
		return false;
	if (!Read32Bit(fs, &m_dirtyFlag))
		return false;
	if (!m_longName.Read(fs))
		return false;;
	if (!m_shortName.Read(fs))
		return false;;
	return true;
}

bool TCategoryEntry::Write(std::ostream *fs) {
	if (!Write32Bit(fs, m_index))
		return false;
	if (!Write32Bit(fs, m_id))
		return false;
	if (!Write32Bit(fs, m_dirtyFlag))
		return false;
	if (!m_longName.Write(fs))
		return false;;
	if (!m_shortName.Write(fs))
		return false;;
	return true;
}

bool TDateBook::Read(std::istream *fs) {
	if (!Read32Bit(fs, &m_versionTag))
		return false;
	if (!m_fileName.Read(fs))
		return false;
	if (!m_tableString.Read(fs))
		return false;
	if (!Read32Bit(fs, &m_nextFree))
		return false;
	if (!Read32Bit(fs, &m_categoryCount))
		return false;
	for (unsigned int i = 0; i < m_categoryCount; i++) {
		TCategoryEntry d;
		if (!d.Read(fs))
			return false;
		m_categoryEntry.push_back(d);
	}

	if (!Read32Bit(fs, &m_resourceID))
		return false;
	if (!Read32Bit(fs, &m_fieldsPerRow))
		return false;
	if (!Read32Bit(fs, &m_recIDPos))
		return false;
	if (!Read32Bit(fs, &m_recStatusPos))
		return false;
	if (!Read32Bit(fs, &m_placementPos))
		return false;
	if (!Read16Bit(fs, &m_fieldCount))
		return false;
	for (unsigned short j = 0; j < m_fieldCount; j++) {
		unsigned short d;
		if (!Read16Bit(fs, &d))
			return false;
		m_fieldEntry.push_back(d);
	}

	if (!Read32Bit(fs, &m_numEntries))
		return false;

	unsigned int actualNumEntries = m_numEntries / 15;

	for (unsigned int k = 0; k < actualNumEntries; k++) {
		TDateBookEntry dbe;
		if (!dbe.Read(fs)) {
			std::cerr << "DateBookEntry[";
			std::cerr << k;
			std::cerr << "] Error\n";
			return false;
		}
		m_dateBookEntry.push_back(dbe);
	}
	return true;
}

bool TDateBook::Write(std::ostream *fs) {
	if (!Write32Bit(fs, m_versionTag))
		return false;
	if (!m_fileName.Write(fs))
		return false;
	if (!m_tableString.Write(fs))
		return false;
	if (!Write32Bit(fs, m_nextFree))
		return false;
	m_categoryCount = (unsigned int)m_categoryEntry.size();
	if (!Write32Bit(fs, m_categoryCount))
		return false;
	for (unsigned int i = 0; i < m_categoryCount; i++) {
		if (!m_categoryEntry[i].Write(fs))
			return false;
	}

	if (!Write32Bit(fs, m_resourceID))
		return false;
	if (!Write32Bit(fs, m_fieldsPerRow))
		return false;
	if (!Write32Bit(fs, m_recIDPos))
		return false;
	if (!Write32Bit(fs, m_recStatusPos))
		return false;
	if (!Write32Bit(fs, m_placementPos))
		return false;
	m_fieldCount = (unsigned int)m_fieldEntry.size();
	if (!Write16Bit(fs, m_fieldCount))
		return false;
	for (unsigned short j = 0; j < m_fieldCount; j++) {
		if (!Write16Bit(fs, m_fieldEntry[j]))
			return false;
	}

	m_numEntries = (unsigned int)(m_dateBookEntry.size() * 15);
	if (!Write32Bit(fs, m_numEntries))
		return false;

	for (int k = 0; k < GetDateBookEntryCount(); k++) {
		if (!GetNthDateBookEntry(k)->Write(fs)) {
			std::cerr << "DateBookEntry[";
			std::cerr << k;
			std::cerr << "] Error\n";
			return false;
		}
	}
	return true;
}

bool TDateBook::Load(std::string path) {
	std::ifstream fs;
	fs.open(path.c_str(), std::ios_base::in | std::ios_base::binary);
	if (!fs.is_open())
		return false;
	bool result = Read(&fs);
	fs.close();
	return result;
}

bool TDateBook::Save(std::string path) {
	std::ofstream fs;
	fs.open(path.c_str(), std::ios_base::out | std::ios_base::binary);
	if (!fs.is_open())
		return false;
	bool result = Write(&fs);
	fs.close();
	return result;
}

int TDateBook::GetDateBookEntryCount() {
	return (int)m_dateBookEntry.size();
	//return (m_numEntries / 15);
}

TDateBookEntry *TDateBook::GetNthDateBookEntry(int n) {
	if (n < 0 || n >= GetDateBookEntryCount())
		return NULL;
	return &m_dateBookEntry[n];
}

}	// namespace
